# flake8: noqa W291
"""Tools for keeping a database of pandora files"""

import os
import sqlite3
import stat

from astropy.time import Time

from .. import CRSOFTVER, LEVEL0_DIR, LEVEL1_DIR, LEVEL2_DIR, __version__, logger
from ..nirda import NIRDALevel1HDUList
from ..utils import get_dpc_hashkey
from ..visda import VISDAFFILevel1HDUList, VISDALevel1HDUList
from .level1 import Level1DataBase
from .mixins import DataBaseMixins


class Level2DataBase(Level1DataBase, DataBaseMixins):
    """Database for managing Level 2 files."""

    def __init__(self):
        self.level = 2
        self.level_dir = LEVEL2_DIR
        self.db_path = f"{self.level_dir}/level{self.level}.db"
        self.conn = sqlite3.connect(self.db_path, timeout=120)
        self.cur = self.conn.cursor()
        self.cur.execute(
            """
        CREATE TABLE IF NOT EXISTS pointings (
            filename TEXT PRIMARY KEY,
            lvlfilename TEXT,
            dir TEXT,
            crsoftver TEXT,
            pfsoftver TEXT,
            finetime INT,
            corstime INT,
            jd FLOAT,
            date STR,
            exptime FLOAT,
            dpc_obs_id INT,
            start FLOAT,
            instrmnt TEXT,
            roisizex INT,
            roisizey INT,
            roistrtx INT,
            roistrty INT,
            next INT,
            astrometry BOOL,
            targ_id STR,
            ra FLOAT,
            dec FLOAT,
            naxis1 INT,
            naxis2 INT,
            naxis3 INT,
            naxis4 INT,
            badchecksum INT,
            baddatasum INT,
            filesize FLOAT
        )
        """
        )
        self.update_str = """INSERT INTO pointings 
        (filename, lvlfilename, dir, crsoftver, pfsoftver, finetime, corstime,
        jd, date, exptime, dpc_obs_id, start, instrmnt, roisizex,
        roisizey, roistrtx, roistrty, next, astrometry,
        targ_id, ra, dec, naxis1, naxis2, naxis3, naxis4, 
        badchecksum, baddatasum, filesize)
        VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)"""
        self.conn.commit()

        os.chmod(
            self.db_path,
            stat.S_IRUSR
            | stat.S_IWUSR  # owner: read/write
            | stat.S_IRGRP
            | stat.S_IWGRP,  # group: read/write
        )
        self.cur.execute(f"ATTACH DATABASE '{LEVEL0_DIR}/pointings.db' AS level0")
        self.cur.execute(f"ATTACH DATABASE '{LEVEL1_DIR}/level1.db' AS level1")

    def process(self, filename):
        logger.info(f"Processing {filename} to Level {self.level}.")
        logger.info("Ensuring file directory present.")
        path = self.get_output_filename(filename)
        os.makedirs("/".join(path.split("/")[:-1]), exist_ok=True)
        if "VisSci" in filename:
            with VISDALevel1HDUList(filename) as hdulist:
                logger.info(f"Converting {filename.split('/')[-1]} to Level 2 product.")
        #         hdulist = getattr(hdulist, f"to_level{self.level}")()
        #         hdulist.writeto(path, overwrite=True, checksum=True)
        if "VisImg" in filename:
            with VISDAFFILevel1HDUList(filename) as hdulist:
                logger.info(f"Converting {filename.split('/')[-1]} to Level 2 product.")
        #         hdulist = getattr(hdulist, f"to_level{self.level}")()
        #         hdulist.writeto(path, overwrite=True, checksum=True)
        if "InfImg" in filename:
            with NIRDALevel1HDUList(filename) as hdulist:
                logger.info(f"Converting {filename.split('/')[-1]} to Level 2 product.")
        #         hdulist = getattr(hdulist, f"to_level{self.level}")()
        #         hdulist.writeto(path, overwrite=True, checksum=True)
        logger.info(f"Wrote {filename.split('/')[-1]} to {path}")
        return self.get_entry(filename)



from . import FORMATSDIR, NIRDAReference
from .fits import PandoraHDUList

__all__ = [
    "NIRDAFFILevel0HDUList",
    "NIRDAFFILevel1HDUList",
    "NIRDALevel0HDUList",
    "NIRDALevel1HDUList",
    "NIRDALevel2HDUList",
]


class NIRDAFFILevel0HDUList(PandoraHDUList):
    filename = FORMATSDIR + "nirda/level0-ffi_nirda.xlsx"
    reference = NIRDAReference

    def to_level1(self):
        return NIRDAFFILevel1HDUList(self)


class NIRDAFFILevel1HDUList(PandoraHDUList):
    filename = FORMATSDIR + "nirda/level1-ffi_nirda.xlsx"
    reference = NIRDAReference

    def to_level1(self):
        raise ValueError("This is a level 1 product.")


class NIRDALevel0HDUList(PandoraHDUList):
    filename = FORMATSDIR + "nirda/level0_nirda.xlsx"
    reference = NIRDAReference

    def to_level1(self):
        return NIRDALevel1HDUList(self)


class NIRDALevel1HDUList(PandoraHDUList):
    filename = FORMATSDIR + "nirda/level1_nirda.xlsx"
    reference = NIRDAReference

    def to_level1(self):
        raise ValueError("This is a level 1 product.")


class NIRDALevel2HDUList(PandoraHDUList):
    filename = FORMATSDIR + "nirda/level2_nirda.xlsx"
    reference = NIRDAReference



# flake8: noqa W291
"""Database tools for MOC files database"""

import os
import sqlite3
import stat
import warnings
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import timedelta
from pathlib import Path

import numpy as np
from astropy.io import fits
from astropy.time import Time

from .. import LEVEL0_DIR, __version__
from .mixins import DataBaseMixins


class Level0DataBase(DataBaseMixins):
    """Database for managing files that have been delivered by MOC."""

    def __init__(self):
        self.db_path = f"{LEVEL0_DIR}/pointings.db"
        self.conn = sqlite3.connect(self.db_path)
        self.cur = self.conn.cursor()

        self.cur.execute(
            """
        CREATE TABLE IF NOT EXISTS pointings (
            filename TEXT PRIMARY KEY,
            dir TEXT,
            crsoftver TEXT,
            pfsoftver TEXT,
            finetime INT,
            corstime INT,
            jd FLOAT,
            date STR,
            exptime FLOAT,
            dpc_obs_id INT,
            start FLOAT,
            instrmnt TEXT,
            roisizex INT,
            roisizey INT,
            roistrtx INT,
            roistrty INT,
            next INT,
            astrometry BOOL,
            targ_id STR,
            ra FLOAT,
            dec FLOAT,
            naxis1 INT,
            naxis2 INT,
            naxis3 INT,
            naxis4 INT,
            badchecksum INT,
            baddatasum INT,
            filesize FLOAT
        )
        """
        )

        self.update_str = """INSERT INTO pointings 
        (filename, dir, crsoftver, pfsoftver, finetime, corstime,
        jd, date, exptime, dpc_obs_id, start, instrmnt, roisizex,
        roisizey, roistrtx, roistrty, next, astrometry,
        targ_id, ra, dec, naxis1, naxis2, naxis3, naxis4, 
        badchecksum, baddatasum, filesize)
        VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)"""
        self.conn.commit()

        os.chmod(
            self.db_path,
            stat.S_IRUSR
            | stat.S_IWUSR  # owner: read/write
            | stat.S_IRGRP
            | stat.S_IWGRP,  # group: read/write
        )

    def __repr__(self):
        return "Pandora Level0DataBase"

    def check_filename_in_database(self, filename):
        self.cur.execute(
            "SELECT 1 FROM pointings WHERE filename=?",
            ((filename.split("/")[-1] if "/" in filename else filename),),
        )
        return self.cur.fetchone() is not None

    def get_entry(self, filename):
        filesize = os.path.getsize(filename) / (1024 * 1024)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")  # capture all warnings

            with fits.open(filename, lazy_load_hdus=True) as hdulist:
                if len(hdulist) <= 1:
                    return
                badchecksum = len(
                    [warn for warn in w if "Checksum" in str(warn.message)]
                )
                baddatasum = len([warn for warn in w if "Datasum" in str(warn.message)])

                hdr = hdulist[0].header
                time = (
                    Time("2000-01-01 12:00:00", scale="tai")
                    + timedelta(
                        seconds=hdr["CORSTIME"],
                        milliseconds=hdr["FINETIME"] / 1e6,
                    )
                ).utc
                hdr1 = hdulist[1].header

                if "FRMTIME" in hdr:
                    frame_time = hdr["FRMTIME"] / 1000
                elif "EXPTIMEU" in hdr:
                    frame_time = (
                        hdr["EXPTIMEU"] * hdr["FRMSCLCT"] / hdr1["NAXIS3"]
                    ) / 1.0e6
                elif "EXPTIME" in hdr:
                    frame_time = (
                        hdr["EXPTIME"] * hdr["FRMSCLCT"] / hdr1["NAXIS3"]
                    ) / 1.0e6

                nframes = hdr1[f"NAXIS{hdr1['NAXIS']}"]
                exptime = nframes * frame_time
                for key in ["FINETIME", "CORSTIME", "INSTRMNT"]:
                    if key not in hdr:
                        return
                return (
                    filename.split("/")[-1],
                    "/".join(filename.split("/")[:-1]),
                    hdr["CRSOFTV"],
                    __version__,
                    hdr["FINETIME"],
                    hdr["CORSTIME"],
                    time.jd,
                    time.isot,
                    exptime,
                    -1,
                    -1,
                    hdr["INSTRMNT"],
                    hdr["ROISIZEX"],
                    hdr["ROISIZEY"],
                    hdr["ROISTRTX"],
                    hdr["ROISTRTY"],
                    len(hdulist),
                    "ASTROMETRY"
                    in np.asarray([hdu.header["extname"] for hdu in hdulist]),
                    hdr["TARG_ID"] if "TARG_ID" in hdr else None,
                    hdr["TARG_RA"] if "TARG_RA" in hdr else None,
                    hdr["TARG_DEC"] if "TARG_DEC" in hdr else None,
                    hdr1["NAXIS1"] if "NAXIS1" in hdr1 else None,
                    hdr1["NAXIS2"] if "NAXIS2" in hdr1 else None,
                    hdr1["NAXIS3"] if "NAXIS3" in hdr1 else None,
                    hdr1["NAXIS4"] if "NAXIS4" in hdr1 else None,
                    badchecksum,
                    baddatasum,
                    filesize,
                )

    def crawl_and_add(self, root):
        for image_type in ["InfImg", "VisSci", "VisImg"]:
            # for path in Path(root).rglob(f"*{image_type}*.fits"):
            #     self.add_entry(self.get_entry(str(path)))
            paths = [
                str(path)
                for path in Path(root).rglob(f"*{image_type}*.fits")
                if not self.check_filename_in_database(str(path))
            ]
            rows = []
            for path in paths:
                values = self.get_entry(path)
                if values is not None:
                    rows.append(values)
            self.add_entries(rows)
        self.update_pointings()

    # def crawl_and_add_parallel(self, root, max_workers=16):
    #     for image_type in ["InfImg", "VisSci", "VisImg"]:
    #         # for path in Path(root).rglob(f"*{image_type}*.fits"):
    #         #     self.add_entry(self.get_entry(str(path)))
    #         paths = [
    #             str(path)
    #             for path in Path(root).rglob(f"*{image_type}*.fits")
    #             if not self.check_filename_in_database(str(path))
    #         ]
    #         rows = []
    #         with ThreadPoolExecutor(max_workers=max_workers) as ex:
    #             futures = [ex.submit(self.get_entry, p) for p in paths]
    #             for fut in as_completed(futures):
    #                 row = fut.result()
    #                 if row is not None:
    #                     rows.append(row)
    #         self.add_entries(rows)
    #     self.update_pointings()

    def _update_dpc_obs_id(self):
        sql = """
        WITH changes AS (
        SELECT
            targ_id,
            CASE
            WHEN targ_id = LAG(targ_id) OVER (ORDER BY jd, targ_id)
            THEN 0          -- same target as previous row → same visit
            ELSE 1          -- target changed (or first row) → new visit
            END AS is_new_visit
        FROM pointings
        ),
        visits AS (
        SELECT
            targ_id,
            SUM(is_new_visit) OVER (ORDER BY jd, targ_id) AS dpc_obs_id
        FROM changes
        )
        UPDATE pointings
        SET dpc_obs_id = (
        SELECT dpc_obs_id FROM visits WHERE visits.targ_id = pointings.targ_id
        );

        """
        self.conn.execute(sql)

    def _update_target(self):
        sql = """
        WITH filled AS (
        SELECT
            targ_id,
            MAX(ra)  OVER (
            PARTITION BY dpc_obs_id
            ORDER BY jd
            ROWS BETWEEN UNBOUNDED PRECEDING AND CURRENT ROW
            ) AS ra_filled,
            MAX(dec) OVER (
            PARTITION BY dpc_obs_id
            ORDER BY jd
            ROWS BETWEEN UNBOUNDED PRECEDING AND CURRENT ROW
            ) AS dec_filled
        FROM pointings
        )
        UPDATE pointings
        SET ra  = (SELECT ra_filled  FROM filled WHERE filled.targ_id = pointings.targ_id),
            dec = (SELECT dec_filled FROM filled WHERE filled.targ_id = pointings.targ_id)
        WHERE ra IS NULL OR dec IS NULL;
        """
        self.conn.execute(sql)

    def _update_start(self):
        sql = """
        WITH filled AS (
        SELECT
            targ_id,
            MIN(jd) OVER (
            PARTITION BY dpc_obs_id
            ORDER BY jd
            ROWS BETWEEN UNBOUNDED PRECEDING AND CURRENT ROW
            ) AS start_filled
        FROM pointings
        )
        UPDATE pointings
        SET start = (SELECT start_filled FROM filled WHERE filled.targ_id = pointings.targ_id)
        """
        self.conn.execute(sql)

    def update_pointings(self):
        self._update_dpc_obs_id()
        self._update_target()
        self._update_start()



"""Submodule for working with file database."""

from .convenience import *  # noqa
from .level0 import Level0DataBase  # noqa
from .level1 import Level1DataBase  # noqa
from .level2 import Level2DataBase  # noqa
from .level3 import Level3DataBase  # noqa



import hashlib
import random
import string
from typing import List

import numpy as np
import openpyxl
import pandas as pd
from astropy.io import fits

BITPIX_DICT = {
    8: (">u1", "Unsigned 8-bit integer, big-endian"),
    16: (">i2", "Signed 16-bit integer, big-endian"),
    32: (">i4", "Signed 32-bit integer, big-endian"),
    -32: (">f4", "32-bit floating-point (float32), big-endian"),
    -64: (">f8", "64-bit floating-point (float64), big-endian"),
}


def get_dpc_hashkey(targ_id, ra, dec):
    key = (targ_id, np.round(ra, 1), np.round(dec, 1))
    return hashlib.sha1(repr(key).encode()).hexdigest()[:8]


def generate_random_table_values(format_code: str, nvalues: int) -> List:
    if format_code.startswith("A"):
        # Character string
        width = int(format_code[1:])
        return [
            "".join(random.choices(string.ascii_letters + string.digits, k=width))
            for n in range(nvalues)
        ]

    elif format_code.startswith("I"):
        # Integer
        width = np.min([int(format_code[1:]), 10])
        return [
            str(random.randint(10 ** (width - 1), 10**width - 1)).rjust(width)
            for n in range(nvalues)
        ]

    elif format_code.startswith("F"):
        # Fixed floating point
        parts = format_code[1:].split(".")
        width = int(parts[0]) - 1
        decimal_places = int(parts[1])
        value = [
            round(
                random.uniform(
                    10 ** (width - decimal_places - 1),
                    10 ** (width - decimal_places) - 1,
                ),
                decimal_places,
            )
            for n in range(nvalues)
        ]
        return [f"{value[n]:>{width}.{decimal_places}f}" for n in range(nvalues)]

    elif format_code.startswith("E") or format_code.startswith("D"):
        # Exponential floating point
        parts = format_code[1:].split(".")
        width = int(parts[0])
        decimal_places = int(parts[1])
        value = random.uniform(
            1e-10, 1e10
        )  # Generating a random number with a large range
        if format_code.startswith("D"):
            return [
                f"{value:>{width}.{decimal_places}E}" for n in range(nvalues)
            ]  # Use 'E' to match exponential format with width
        else:
            return [
                f"{value:>{width}.{decimal_places}e}" for n in range(nvalues)
            ]  # 'e' is used for lowercase exponent notation

    else:
        raise ValueError("Unsupported format code")


def get_excel_sheet(fname, extno=0):
    wb = openpyxl.load_workbook(fname, data_only=False)
    ws = wb.worksheets[extno]

    rows = []
    for row in ws.iter_rows(values_only=False):
        rows.append(
            [
                (
                    cell.value
                    if cell.data_type != "n"
                    else cell.number_format and cell._value
                )
                for cell in row
            ]
        )

    df = pd.DataFrame(rows)
    df.columns = df.loc[0]
    df = df[1:].reset_index(drop=True)
    df = df[~df.apply(lambda row: row.isnull().all(), axis=1)].reset_index(drop=True)
    return df


def hdulist_to_excel(hdulist, filename="output.xlsx"):
    if isinstance(hdulist, str):
        hdulist = fits.open(hdulist)

    sheet_names = ["Primary", *[f"Ext{idx}" for idx in np.arange(1, len(hdulist))]]
    dfs = pd.DataFrame(
        [
            np.asarray([hdu.header["EXTNAME"] for hdu in hdulist]),
            np.asarray([type(hdu).__name__ for hdu in hdulist]),
            np.zeros(len(hdulist), bool),
        ]
    ).T
    dfs.columns = ["Extension", "Type", "Optional"]
    dfs = [
        dfs,
        *[
            pd.DataFrame(
                np.asarray([np.asarray(card) for card in hdu.header.cards]),
                columns=["Name", "Example Value", "Comment"],
            )
            for hdu in hdulist
        ],
    ]
    with pd.ExcelWriter(filename, engine="openpyxl") as writer:
        dfs[0].to_excel(writer, sheet_name="Structure", index=False)
        for name, df in zip(sheet_names, dfs[1:]):
            df["Fixed Value"] = None
            df[["Name", "Fixed Value", "Example Value", "Comment"]].to_excel(
                writer, sheet_name=name, index=False
            )



# flake8: noqa W291
"""Tools for keeping a database of pandora files"""

import os
import sqlite3
import stat
import warnings
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import timedelta
from pathlib import Path

import numpy as np
import pandas as pd
from astropy.io import fits
from astropy.time import Time

from . import CRSOFTVER, DATA_DIR, LEVEL0_DIR, LEVEL1_DIR, __version__, logger
from .nirda import NIRDALevel0HDUList
from .utils import get_dpc_hashkey
from .visda import VISDAFFILevel0HDUList, VISDALevel0HDUList


def update_filedatabase() -> None:
    """
    Creates and updates to the SQLite database file.
    """
    with FileDataBase() as db:
        db.crawl_and_add_parallel(DATA_DIR)
        db.update_pointings()


def delete_filedatabase() -> None:
    """
    Deletes the SQLite database file.

    Raises
    ------
    FileNotFoundError
        If the database file does not exist.
    """
    db_path = f"{LEVEL0_DIR}/pointings.db"
    if os.path.exists(db_path):
        os.remove(db_path)
        logger.warning(f"Database at {db_path} has been deleted.")
    else:
        logger.warning(f"Tried to delete FileDataBase. No database found at {db_path}.")


def delete_level1database() -> None:
    """
    Deletes the SQLite database file.

    Raises
    ------
    FileNotFoundError
        If the database file does not exist.
    """
    db_path = f"{LEVEL1_DIR}/level1.db"
    if os.path.exists(db_path):
        os.remove(db_path)
        logger.warning(f"Database at {db_path} has been deleted.")
    else:
        logger.warning(
            f"Tried to delete Level1DataBase. No database found at {db_path}."
        )


def clean_level1database(crsoftver=CRSOFTVER, pfsoftver=__version__):
    """Be warned do NOT run this on multiple nodes, run on one node!"""
    logger.info("Running `clean_level1database`")
    with Level1DataBase() as l1db:
        for file in l1db.files_to_process(crsoftver=crsoftver):
            l1db.check_file_in_database(
                file, crsoftver=crsoftver, pfsoftver=pfsoftver, remove_bad_entries=True
            )
    logger.info("Finished `clean_level1database`")


class DataBaseMixins:
    def to_pandas(self, time_range=None, targ_id=None, dpc_obs_id=None):
        sql = "SELECT * FROM pointings"
        params = []

        where_clauses = []

        if time_range is not None:
            start, end = _process_time(time_range[0]), _process_time(time_range[1])
            start, end = np.sort([start, end])
            where_clauses.append("jd BETWEEN ? AND ?")
            params.extend([start, end])

        if targ_id is not None:
            where_clauses.append("targ_id = ?")
            params.append(targ_id)

        if dpc_obs_id is not None:
            where_clauses.append("dpc_obs_id = ?")
            params.append(dpc_obs_id)

        if where_clauses:
            sql += " WHERE " + " AND ".join(where_clauses)

        return pd.read_sql_query(sql, self.conn, params=params)

    def add_entry(self, values):
        self.cur.execute(
            self.update_str,
            values,
        )
        self.conn.commit()

    def add_entries(self, values):
        self.cur.executemany(
            self.update_str,
            values,
        )
        self.conn.commit()


class FileDataBase(DataBaseMixins):
    """Database for managing files that have been delivered by MOC."""

    def __init__(self):
        self.db_path = f"{LEVEL0_DIR}/pointings.db"
        self.conn = sqlite3.connect(self.db_path)
        self.cur = self.conn.cursor()

        self.cur.execute(
            """
        CREATE TABLE IF NOT EXISTS pointings (
            filename TEXT PRIMARY KEY,
            dir TEXT,
            crsoftver TEXT,
            pfsoftver TEXT,
            finetime INT,
            corstime INT,
            jd FLOAT,
            date STR,
            dpc_obs_id INT,
            start FLOAT,
            instrmnt TEXT,
            roisizex INT,
            roisizey INT,
            roistrtx INT,
            roistrty INT,
            next INT,
            astrometry BOOL,
            targ_id STR,
            ra FLOAT,
            dec FLOAT,
            naxis1 INT,
            naxis2 INT,
            naxis3 INT,
            naxis4 INT,
            badchecksum INT,
            baddatasum INT,
            filesize FLOAT
        )
        """
        )

        self.update_str = """INSERT INTO pointings 
        (filename, dir, crsoftver, pfsoftver, finetime, corstime,
        jd, date, dpc_obs_id, start, instrmnt, roisizex,
        roisizey, roistrtx, roistrty, next, astrometry,
        targ_id, ra, dec, naxis1, naxis2, naxis3, naxis4, 
        badchecksum, baddatasum, filesize)
        VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)"""
        self.conn.commit()

        os.chmod(
            self.db_path,
            stat.S_IRUSR
            | stat.S_IWUSR  # owner: read/write
            | stat.S_IRGRP
            | stat.S_IWGRP,  # group: read/write
        )

    def __repr__(self):
        return "Pandora FileDataBase"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        """Closes the database connection."""
        if self.conn:
            self.conn.close()

    def check_filename_in_database(self, filename):
        self.cur.execute(
            "SELECT 1 FROM pointings WHERE filename=?",
            ((filename.split("/")[-1] if "/" in filename else filename),),
        )
        return self.cur.fetchone() is not None

    def get_entry(self, filename):
        filesize = os.path.getsize(filename) / (1024 * 1024)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")  # capture all warnings

            with fits.open(filename, lazy_load_hdus=True) as hdulist:
                if len(hdulist) <= 1:
                    return
                badchecksum = len(
                    [warn for warn in w if "Checksum" in str(warn.message)]
                )
                baddatasum = len([warn for warn in w if "Datasum" in str(warn.message)])

                hdr = hdulist[0].header
                time = (
                    Time("2000-01-01 12:00:00", scale="tai")
                    + timedelta(
                        seconds=hdr["CORSTIME"],
                        milliseconds=hdr["FINETIME"] / 1e6,
                    )
                ).utc
                hdr1 = hdulist[1].header
                for key in ["FINETIME", "CORSTIME", "INSTRMNT"]:
                    if key not in hdr:
                        return
                return (
                    filename.split("/")[-1],
                    "/".join(filename.split("/")[:-1]),
                    hdr["CRSOFTV"],
                    __version__,
                    hdr["FINETIME"],
                    hdr["CORSTIME"],
                    time.jd,
                    time.isot,
                    -1,
                    -1,
                    hdr["INSTRMNT"][0],
                    hdr["ROISIZEX"],
                    hdr["ROISIZEY"],
                    hdr["ROISTRTX"],
                    hdr["ROISTRTY"],
                    len(hdulist),
                    "ASTROMETRY"
                    in np.asarray([hdu.header["extname"] for hdu in hdulist]),
                    hdr["TARG_ID"] if "TARG_ID" in hdr else None,
                    hdr["TARG_RA"] if "TARG_RA" in hdr else None,
                    hdr["TARG_DEC"] if "TARG_DEC" in hdr else None,
                    hdr1["NAXIS1"] if "NAXIS1" in hdr1 else None,
                    hdr1["NAXIS2"] if "NAXIS2" in hdr1 else None,
                    hdr1["NAXIS3"] if "NAXIS3" in hdr1 else None,
                    hdr1["NAXIS4"] if "NAXIS4" in hdr1 else None,
                    badchecksum,
                    baddatasum,
                    filesize,
                )

    def crawl_and_add(self, root):
        for image_type in ["InfImg", "VisSci", "VisImg"]:
            # for path in Path(root).rglob(f"*{image_type}*.fits"):
            #     self.add_entry(self.get_entry(str(path)))
            self.add_entries(
                [
                    self.get_entry(str(path))
                    for path in Path(root).rglob(f"*{image_type}*.fits")
                    if not self.check_filename_in_database(str(path))
                ]
            )

    def crawl_and_add_parallel(self, root, max_workers=16):
        for image_type in ["InfImg", "VisSci", "VisImg"]:
            # for path in Path(root).rglob(f"*{image_type}*.fits"):
            #     self.add_entry(self.get_entry(str(path)))
            paths = [
                str(path)
                for path in Path(root).rglob(f"*{image_type}*.fits")
                if not self.check_filename_in_database(str(path))
            ]
            rows = []
            with ThreadPoolExecutor(max_workers=max_workers) as ex:
                futures = [ex.submit(self.get_entry, p) for p in paths]
                for fut in as_completed(futures):
                    row = fut.result()
                    if row is not None:
                        rows.append(row)
            self.add_entries(rows)

    def _update_dpc_obs_id(self):
        sql = """
        WITH changes AS (
        SELECT
            targ_id,
            CASE
            WHEN targ_id = LAG(targ_id) OVER (ORDER BY jd, targ_id)
            THEN 0          -- same target as previous row → same visit
            ELSE 1          -- target changed (or first row) → new visit
            END AS is_new_visit
        FROM pointings
        ),
        visits AS (
        SELECT
            targ_id,
            SUM(is_new_visit) OVER (ORDER BY jd, targ_id) AS dpc_obs_id
        FROM changes
        )
        UPDATE pointings
        SET dpc_obs_id = (
        SELECT dpc_obs_id FROM visits WHERE visits.targ_id = pointings.targ_id
        );

        """
        self.conn.execute(sql)

    def _update_target(self):
        sql = """
        WITH filled AS (
        SELECT
            targ_id,
            MAX(ra)  OVER (
            PARTITION BY dpc_obs_id
            ORDER BY jd
            ROWS BETWEEN UNBOUNDED PRECEDING AND CURRENT ROW
            ) AS ra_filled,
            MAX(dec) OVER (
            PARTITION BY dpc_obs_id
            ORDER BY jd
            ROWS BETWEEN UNBOUNDED PRECEDING AND CURRENT ROW
            ) AS dec_filled
        FROM pointings
        )
        UPDATE pointings
        SET ra  = (SELECT ra_filled  FROM filled WHERE filled.targ_id = pointings.targ_id),
            dec = (SELECT dec_filled FROM filled WHERE filled.targ_id = pointings.targ_id)
        WHERE ra IS NULL OR dec IS NULL;
        """
        self.conn.execute(sql)

    def _update_start(self):
        sql = """
        WITH filled AS (
        SELECT
            targ_id,
            MIN(jd) OVER (
            PARTITION BY dpc_obs_id
            ORDER BY jd
            ROWS BETWEEN UNBOUNDED PRECEDING AND CURRENT ROW
            ) AS start_filled
        FROM pointings
        )
        UPDATE pointings
        SET start = (SELECT start_filled FROM filled WHERE filled.targ_id = pointings.targ_id)
        """
        self.conn.execute(sql)

    def update_pointings(self):
        self._update_dpc_obs_id()
        self._update_target()
        self._update_start()


class Level1DataBase(DataBaseMixins):
    """Database for managing Level 1 files."""

    def __init__(self):
        self.db_path = f"{LEVEL1_DIR}/level1.db"
        self.conn = sqlite3.connect(self.db_path)
        self.cur = self.conn.cursor()
        self.cur.execute(
            """
        CREATE TABLE IF NOT EXISTS pointings (
            filename TEXT PRIMARY KEY,
            dir TEXT,
            crsoftver TEXT,
            pfsoftver TEXT,
            finetime INT,
            corstime INT,
            jd FLOAT,
            date STR,
            dpc_obs_id INT,
            start FLOAT,
            instrmnt TEXT,
            roisizex INT,
            roisizey INT,
            roistrtx INT,
            roistrty INT,
            next INT,
            astrometry BOOL,
            targ_id STR,
            ra FLOAT,
            dec FLOAT,
            naxis1 INT,
            naxis2 INT,
            naxis3 INT,
            naxis4 INT,
            badchecksum INT,
            baddatasum INT,
            filesize FLOAT
        )
        """
        )
        self.update_str = """INSERT INTO pointings 
        (filename, dir, crsoftver, pfsoftver, finetime, corstime,
        jd, date, dpc_obs_id, start, instrmnt, roisizex,
        roisizey, roistrtx, roistrty, next, astrometry,
        targ_id, ra, dec, naxis1, naxis2, naxis3, naxis4, 
        badchecksum, baddatasum, filesize)
        VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)"""
        self.conn.commit()

        os.chmod(
            self.db_path,
            stat.S_IRUSR
            | stat.S_IWUSR  # owner: read/write
            | stat.S_IRGRP
            | stat.S_IWGRP,  # group: read/write
        )
        # Add the level0 database in
        self.cur.execute(f"ATTACH DATABASE '{LEVEL0_DIR}/pointings.db' AS level0")

    def __repr__(self):
        return "Pandora Level1DataBase"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        """Closes the database connection."""
        if self.conn:
            self.conn.close()

    def _is_bad_value(
        self, fname, expected_entry, entry, name, level, remove_bad_entries=True
    ):
        if expected_entry != entry:
            logger.warning(
                f"File {fname} exists in Level {level} database with `{name}` set to `{entry}` requested value was `{expected_entry}`."
            )
            if remove_bad_entries:
                if level == 0:
                    self.cur.execute(
                        "DELETE FROM level0.pointings WHERE filename=?", (fname,)
                    )
                if level == 1:
                    self.cur.execute("DELETE FROM pointings WHERE filename=?", (fname,))
                self.conn.commit()
                logger.warning(f"File {fname} removed from Level {level} database.")
            return True
        return False

    def _is_bad_row(
        self,
        filename,
        crsoftver=CRSOFTVER,
        pfsoftver=__version__,
        badchecksum=0,
        baddatasum=0,
        level=0,
        remove_bad_entries=False,
    ):
        fname = filename.split("/")[-1] if "/" in filename else filename
        self.cur.execute(
            f"SELECT crsoftver, pfsoftver, badchecksum, baddatasum FROM {'level0.' if level == 0 else ''}pointings WHERE filename=?",
            (fname,),
        )
        row = self.cur.fetchone()

        if row is None:
            return False
        (
            database_crsoftver,
            database_pfsoftver,
            database_badchecksum,
            database_baddatasum,
        ) = row
        is_bad_row = False
        # Log any discrepancies
        if crsoftver is not None:
            is_bad_row |= self._is_bad_value(
                fname=fname,
                expected_entry=crsoftver,
                entry=database_crsoftver,
                name="CRSOFTV",
                level=0,
                remove_bad_entries=remove_bad_entries,
            )
        if pfsoftver is not None:
            if database_pfsoftver is not None:
                is_bad_row |= self._is_bad_value(
                    fname=fname,
                    expected_entry=pfsoftver,
                    entry=database_pfsoftver,
                    name="PFSOFTV",
                    level=0,
                    remove_bad_entries=remove_bad_entries,
                )
        if badchecksum is not None:
            is_bad_row |= self._is_bad_value(
                fname=fname,
                expected_entry=badchecksum,
                entry=database_badchecksum,
                name="badchecksum",
                level=0,
                remove_bad_entries=remove_bad_entries,
            )
        if baddatasum is not None:
            is_bad_row |= self._is_bad_value(
                fname=fname,
                expected_entry=baddatasum,
                entry=database_baddatasum,
                name="baddatasum",
                level=0,
                remove_bad_entries=remove_bad_entries,
            )
        return is_bad_row

    def check_filename_in_database(self, filename, level=0):
        fname = filename.split("/")[-1] if "/" in filename else filename
        self.cur.execute(
            f"SELECT crsoftver, badchecksum, baddatasum FROM {'level0.' if level == 0 else ''}pointings WHERE filename=?",
            (fname,),
        )
        return self.cur.fetchone() is not None

    # def _check_entry(
    #     self, fname, expected_entry, entry, name, level, remove_bad_entries=True
    # ):
    #     if expected_entry != entry:
    #         logger.warning(
    #             f"File {fname} exists in Level {level} database with `{name}` set to `{entry}` requested value was `{expected_entry}`."
    #         )
    #         if remove_bad_entries:
    #             if level == 0:
    #                 self.cur.execute(
    #                     "DELETE FROM level0.pointings WHERE filename=?", (fname,)
    #                 )
    #             if level == 1:
    #                 self.cur.execute("DELETE FROM pointings WHERE filename=?", (fname,))
    #             self.conn.commit()
    #             logger.warning(f"File {fname} removed from Level {level} database.")

    # def check_file_in_level0_database(
    #     self,
    #     filename,
    #     crsoftver=None,
    #     badchecksum=0,
    #     baddatasum=0,
    #     remove_bad_entries=False,
    # ):
    #     fname = filename.split("/")[-1] if "/" in filename else filename
    #     self.cur.execute(
    #         "SELECT crsoftver, badchecksum, baddatasum FROM level0.pointings WHERE filename=?",
    #         (fname,),
    #     )
    #     row = self.cur.fetchone()

    #     if row is None:
    #         return False

    #     database_crsoftver, database_badchecksum, database_baddatasum = row

    #     # Log any discrepancies
    #     if crsoftver is not None:
    #         self._check_entry(
    #             fname=fname,
    #             expected_entry=crsoftver,
    #             entry=database_crsoftver,
    #             name="CRSOFTV",
    #             level=0,
    #             remove_bad_entries=remove_bad_entries,
    #         )
    #     if badchecksum is not None:
    #         self._check_entry(
    #             fname=fname,
    #             expected_entry=badchecksum,
    #             entry=database_badchecksum,
    #             name="badchecksum",
    #             level=0,
    #             remove_bad_entries=remove_bad_entries,
    #         )
    #     if baddatasum is not None:
    #         self._check_entry(
    #             fname=fname,
    #             expected_entry=baddatasum,
    #             entry=database_baddatasum,
    #             name="baddatasum",
    #             level=0,
    #             remove_bad_entries=remove_bad_entries,
    #         )
    #     if remove_bad_entries:
    #         return False
    #     return True

    # def check_file_in_database(
    #     self,
    #     filename,
    #     crsoftver=None,
    #     pfsoftver=None,
    #     badchecksum=0,
    #     baddatasum=0,
    #     remove_bad_entries=False,
    # ):
    #     fname = filename.split("/")[-1] if "/" in filename else filename
    #     self.cur.execute(
    #         "SELECT crsoftver, pfsoftver, badchecksum, baddatasum FROM pointings WHERE filename=?",
    #         (fname,),
    #     )
    #     row = self.cur.fetchone()

    #     if row is None:
    #         return False
    #     (
    #         database_crsoftver,
    #         database_pfsoftver,
    #         database_badchecksum,
    #         database_baddatasum,
    #     ) = row
    #     # Log any discrepancies
    #     if crsoftver is not None:
    #         self._check_entry(
    #             fname=fname,
    #             expected_entry=crsoftver,
    #             entry=database_crsoftver,
    #             name="CRSOFTV",
    #             level=1,
    #             remove_bad_entries=remove_bad_entries,
    #         )
    #     if pfsoftver is not None:
    #         self._check_entry(
    #             fname=fname,
    #             expected_entry=pfsoftver,
    #             entry=database_pfsoftver,
    #             name="PFSOFTV",
    #             level=1,
    #             remove_bad_entries=remove_bad_entries,
    #         )
    #     if badchecksum is not None:
    #         self._check_entry(
    #             fname=fname,
    #             expected_entry=badchecksum,
    #             entry=database_badchecksum,
    #             name="badchecksum",
    #             level=1,
    #             remove_bad_entries=remove_bad_entries,
    #         )
    #     if baddatasum is not None:
    #         self._check_entry(
    #             fname=fname,
    #             expected_entry=baddatasum,
    #             entry=database_baddatasum,
    #             name="baddatasum",
    #             level=1,
    #             remove_bad_entries=remove_bad_entries,
    #         )
    #     if remove_bad_entries:
    #         return False
    #     return True

    def files_to_process(self, crsoftver=CRSOFTVER):
        cur = self.conn.cursor()
        cur.execute("SELECT filename, dir FROM level0.pointings")

        row = cur.fetchone()
        while row is not None:
            if (not self.check_filename_in_database(row[0], level=1)) & (
                not self._is_bad_row(row[0], crsoftver=crsoftver, level=0)
            ):
                yield (f"{row[1]}/{row[0]}")
            row = cur.fetchone()

    def get_entry(self, filename):
        fname = filename.split("/")[-1] if "/" in filename else filename
        self.cur.execute(
            "SELECT * FROM level0.pointings WHERE filename=?",
            (fname,),
        )
        row = self.cur.fetchone()
        return (*row[:3], __version__, *row[3:])

    def get_level1_filename(self, filename):
        fname = filename.split("/")[-1] if "/" in filename else filename
        self.cur.execute(
            "SELECT start, targ_id, ra, dec FROM level0.pointings WHERE filename=?",
            (fname,),
        )
        row = self.cur.fetchone()
        if row is None:
            return None
        t = Time(row[0], format="jd").to_datetime()

        return f"{LEVEL1_DIR}/{t.year}/{t.month}/{t.day}/{get_dpc_hashkey(*row[1:])}/{fname}"

    def process(self, filename):
        logger.info(f"Processing {filename} to Level 1.")
        if "VisSci" in filename:
            with VISDALevel0HDUList(filename) as hdulist:
                path = self.get_level1_filename(filename)
                hdulist = hdulist.to_level1()
                # hdulist.writeto(path, overwrite=True, checksum=True)
                logger.info(f"Wrote {filename} to {path}")
        if "VisImg" in filename:
            with VISDAFFILevel0HDUList(filename) as hdulist:
                path = self.get_level1_filename(filename)
                hdulist = hdulist.to_level1()
                # hdulist.writeto(path, overwrite=True, checksum=True)
                logger.info(f"Wrote {filename} to {path}")
        if "InfImg" in filename:
            with NIRDALevel0HDUList(filename) as hdulist:
                path = self.get_level1_filename(filename)
                hdulist = hdulist.to_level1()
                # hdulist.writeto(path, overwrite=True, checksum=True)
                logger.info(f"Wrote {filename} to {path}")

        return self.get_entry(filename)

    def crawl_and_process_parallel(self, max_workers=16, crsoftver=CRSOFTVER):
        paths = self.files_to_process(crsoftver=crsoftver)
        rows = []
        with ThreadPoolExecutor(max_workers=max_workers) as ex:
            futures = [ex.submit(self.process, p) for p in paths]
            for fut in as_completed(futures):
                row = fut.result()
                if row is not None:
                    rows.append(row)
        self.add_entries(rows)

    def crawl_and_process(self, crsoftver=None):
        for path in self.files_to_process(crsoftver=CRSOFTVER):
            self.add_entry(self.process(path))


def _process_time(time):
    if isinstance(time, Time):
        time = time.jd
    elif not isinstance(time, float):
        try:
            time = Time(time).jd
        except ValueError:
            raise ValueError("`time_range` must be in JD.")
    return time



"""Class to handle Pandora fits files"""

import numpy as np

# import pandas as pd
from astropy.io import fits

from . import logger
from .processing import ProcessingMixins
from .utils import BITPIX_DICT, generate_random_table_values, get_excel_sheet

__all__ = ["FITSTemplateException", "FITSValueException", "PandoraHDUList"]


class FITSTemplateException(Exception):
    """Custom exception for fits files not having the right shape."""

    def __init__(self, message):
        super().__init__(message)


class FITSValueException(Exception):
    """Custom exception for fits files not having the right values."""

    def __init__(self, message):
        super().__init__(message)


def _clean_header_cards(hdr: fits.Header):
    """Cleans the list of cards to ensure they have reasonable values."""
    for key in hdr:
        if hdr[key] in ["TRUE", "True", "T"]:
            hdr[key] = True
        if hdr[key] in ["FALSE", "False", "F"]:
            hdr[key] = False

        def to_number(x):
            if isinstance(x, str):
                try:
                    return int(x.strip())
                except (ValueError, TypeError):
                    try:
                        return float(x.strip())
                    except (ValueError, TypeError):
                        return x.strip()
            return x

        hdr[key] = to_number(hdr[key])
    return hdr


class PandoraHDUList(fits.HDUList, ProcessingMixins):
    """Base class, not designed to be used. Adds mixins to the fits.HDUList object"""

    def _get_default_cards(self, extname):
        if isinstance(extname, int):
            extname = self.extension_names[extname.lower()]
        return [
            fits.Card(d.iloc[0], d.iloc[1] if d.iloc[1] != "" else d.iloc[2], d.iloc[3])
            for _, d in self.extension_headers[extname.lower()].fillna("").iterrows()
        ]

    def _get_mandetory_cards(self, extname):
        if isinstance(extname, int):
            extname = self.extension_names[extname.lower()]
        return [
            fits.Card(d.iloc[0], d.iloc[1], d.iloc[3])
            for _, d in self.extension_headers[extname.lower()].fillna("").iterrows()
        ]

    def _validate_ext_types(self):
        """Validate that the extensions have the correct types, e.g. ImageHDU, TableHDU, etc"""
        for hdu, expected_type in zip(self, self.extension_types):
            if not isinstance(hdu, getattr(fits, expected_type)):
                raise FITSTemplateException(
                    f"Data doesn't match format for {self.__class__.__name__}. Expected extension type {expected_type}, got {hdu}."
                )

    def _validate_n_ext(self):
        """Validate that all the necessary extensions are present."""
        k = np.in1d(
            self.extension_names, [hdu.header["EXTNAME"].lower() for hdu in self]
        )
        for name in self.extension_names[~k]:
            if not self.structure[
                self.structure.Extension.str.lower() == name
            ].Optional.values:
                raise FITSTemplateException(
                    f"Data doesn't match format for {self.__class__.__name__}. Expected extension {name}, but none found."
                )
        self.extension_headers = {
            n: self.extension_headers[n] for n in self.extension_names[k]
        }
        self.extension_types = self.extension_types[k]
        self.extension_names = self.extension_names[k]

    def _validate_data(self):
        """Check the data in the fits file is all the right dtype, given expected `bitpix`"""
        for extname, hdu in zip(self.extension_names, self):
            if hdu.header["EXTNAME"] == "PRIMARY":
                continue
            if isinstance(hdu, fits.ImageHDU):
                expected_header = fits.Header(self._get_mandetory_cards(extname))
                expected_type, expected_type_str = BITPIX_DICT[
                    int(expected_header["bitpix"])
                ]
                if not hdu.data.dtype == expected_type:
                    if int(expected_header["bitpix"]) == 32:
                        if not (int(hdu.header["BSCALE"]) == 1) & (
                            int(hdu.header["BZERO"]) == 2**31
                        ):
                            raise FITSTemplateException(
                                f"Data doesn't match format for {self.__class__.__name__}."
                                f" Expected data type of np.uint32, got {hdu.data.dtype}"
                            )
                    else:
                        raise FITSTemplateException(
                            f"Data doesn't match format for {self.__class__.__name__}. "
                            f"Expected data of type {expected_type} ({expected_type_str}), got {hdu.data.dtype}"
                        )

    def _validate_mandetory_headers(self, warn=False):
        """Validate the data contains mandetory header cards."""
        for extname, hdu in zip(self.extension_names, self):
            hdr = _clean_header_cards(hdu.header)
            expected_header = _clean_header_cards(
                fits.Header(self._get_mandetory_cards(extname))
            )
            for key in expected_header:
                # fill missing cards
                if key not in hdr:
                    if warn:
                        hdr[key] = expected_header[key]
                        logger.warning(
                            f"Key {key} expected in extension `{extname}` but not found. Added this key."
                        )
                    else:
                        raise FITSValueException(
                            f"{key} header keyword expected for {self.__class__.__name__} in extension `{extname}`,"
                            " but not found in data provided."
                        )
                # check mandetory cards have the correct values
                if expected_header[key] not in ["", None, np.nan]:
                    if hdr[key] != expected_header[key]:
                        if isinstance(expected_header[key], bool):
                            continue
                        raise FITSValueException(
                            f"{key} expected to have value of {expected_header[key]}, but has value {hdr[key]}."
                        )

    def _validate_no_extra_keywords(self, warn=False):
        """Validate there are no additional keywords in the file"""
        for extname, hdu in zip(self.extension_names, self):
            hdr = _clean_header_cards(hdu.header)
            expected_header = _clean_header_cards(
                fits.Header(self._get_mandetory_cards(extname))
            )
            for key in hdr:
                if key not in expected_header:
                    if warn:
                        hdr.pop(key)
                        logger.warning(
                            f"Key {key} found in extension `{extname}` but not expected. Removed this key."
                        )
                    else:
                        raise FITSTemplateException(
                            f"{key} header keyword is not expected for {self.__class__.__name__} in extension `{extname}`."
                        )

    def _get_dummy_hdus(self):
        hdulist = []
        for extname, exttype in zip(self.extension_names, self.extension_types):
            cards = self._get_default_cards(extname)
            hdr = _clean_header_cards(fits.Header(cards))
            data = None
            if exttype == "PrimaryHDU":
                hdu = fits.PrimaryHDU(header=hdr)
            elif exttype == "CompImageHDU":
                shape = tuple(
                    [
                        int(hdr[f"NAXIS{naxis}"])
                        for naxis in np.arange(1, hdr["NAXIS"] + 1)[::-1]
                    ]
                )
                if "BSCALE" in hdr:
                    if (int(hdr["BSCALE"]) == 1) and (int(hdr["BZERO"]) == 2**15):
                        data = np.ones(shape, dtype=np.uint16)
                    else:
                        raise FITSValueException("Can not parse data type")
                else:
                    data = np.ones(shape, dtype=BITPIX_DICT[hdr["BITPIX"]][0])
                hdu = fits.CompImageHDU(header=hdr, data=data)
            elif exttype == "ImageHDU":
                shape = tuple(
                    [
                        int(hdr[f"NAXIS{naxis}"])
                        for naxis in np.arange(1, hdr["NAXIS"] + 1)[::-1]
                    ]
                )
                if "BSCALE" in hdr:
                    if (int(hdr["BSCALE"]) == 1) and (int(hdr["BZERO"]) == 2**31):
                        data = np.ones(shape, dtype=np.uint32)
                    else:
                        raise FITSValueException("Can not parse data type")
                else:
                    data = np.ones(shape, dtype=BITPIX_DICT[hdr["BITPIX"]][0])
                hdu = fits.ImageHDU(header=hdr, data=data)
            elif exttype == "TableHDU":
                ncolumns = len(
                    [c.keyword for c in cards if c.keyword.startswith("TTYPE")]
                )
                columns = [
                    fits.Column(
                        name=hdr[f"TTYPE{idx}"],
                        format=hdr[f"TFORM{idx}"],
                        unit=hdr[f"TUNIT{idx}"] if f"TUNIT{idx}" in hdr else "",
                        array=(
                            generate_random_table_values(
                                hdr[f"TFORM{idx}"], hdr["NAXIS2"]
                            )
                            if hdr["NAXIS2"] != ""
                            else None
                        ),
                    )
                    for idx in np.arange(1, ncolumns + 1)
                ]

                hdu = fits.TableHDU.from_columns(columns, header=hdr)
            else:
                raise FITSValueException(f"No extension type {exttype}.")
            cards = self._get_mandetory_cards(extname)
            _ = [hdu.header.append(card) for card in cards if card[0] not in hdu.header]
            hdulist.append(hdu)
        return fits.HDUList(hdulist)

    def __init__(self, file=None, validate=True, warn=False):
        self.warn = warn
        self.structure = get_excel_sheet(self.filename, 0)
        # load in header formats
        self.extension_headers = [
            get_excel_sheet(self.filename, idx + 1)
            for idx in range(len(self.structure))
        ]
        # convert to dictionary
        self.extension_headers = {
            h["Fixed Value"][h.Name.isin(["EXTNAME"])].values[0].lower(): h
            for h in self.extension_headers
        }
        # remove optional extensions
        self.structure = self.structure[
            self.structure["Extension"].str.lower().isin(self.extension_headers.keys())
        ]
        self.extension_names = np.asarray(self.structure.Extension.str.lower().values)
        self.extension_types = np.asarray(self.structure.Type.values)

        if file is None:
            hdulist = self._get_dummy_hdus()
            super().__init__(hdulist)
        elif isinstance(file, str):
            super().__init__(fits.open(file))
        elif isinstance(file, fits.HDUList):
            super().__init__(file)
        if validate:
            self._validate_n_ext()
            self._validate_ext_types()
            self._validate_data()
            self._validate_mandetory_headers(warn=self.warn)
            self._validate_no_extra_keywords(warn=self.warn)

    def writeto(
        self,
        fileobj,
        output_verify="exception",
        overwrite=False,
        checksum=False,
    ):
        self.__class__(
            fits.HDUList(self).writeto(
                fileobj=fileobj,
                output_verify=output_verify,
                overwrite=overwrite,
                checksum=checksum,
            )
        )



# flake8: noqa W291
"""Tools for keeping a database of pandora files"""

import os
import sqlite3
import stat

from astropy.time import Time

from .. import (
    CRSOFTVER,
    LEVEL0_DIR,
    LEVEL1_DIR,
    LEVEL2_DIR,
    LEVEL3_DIR,
    __version__,
    logger,
)
from ..nirda import NIRDALevel0HDUList
from ..utils import get_dpc_hashkey
from ..visda import VISDAFFILevel0HDUList, VISDALevel0HDUList
from .level1 import Level1DataBase
from .mixins import DataBaseMixins


class Level3DataBase(Level1DataBase, DataBaseMixins):
    """Database for managing Level 2 files."""

    def __init__(self):
        self.level = 3
        self.level_dir = LEVEL3_DIR
        self.db_path = f"{self.level_dir}/level{self.level}.db"
        self.conn = sqlite3.connect(self.db_path, timeout=120)
        self.cur = self.conn.cursor()
        self.cur.execute(
            """
        CREATE TABLE IF NOT EXISTS pointings (
            filename TEXT PRIMARY KEY,
            lvlfilename TEXT,
            dir TEXT,
            crsoftver TEXT,
            pfsoftver TEXT,
            finetime INT,
            corstime INT,
            jd FLOAT,
            date STR,
            exptime FLOAT,
            dpc_obs_id INT,
            start FLOAT,
            instrmnt TEXT,
            roisizex INT,
            roisizey INT,
            roistrtx INT,
            roistrty INT,
            next INT,
            astrometry BOOL,
            targ_id STR,
            ra FLOAT,
            dec FLOAT,
            naxis1 INT,
            naxis2 INT,
            naxis3 INT,
            naxis4 INT,
            badchecksum INT,
            baddatasum INT,
            filesize FLOAT
        )
        """
        )
        self.update_str = """INSERT INTO pointings 
        (filename, lvlfilename, dir, crsoftver, pfsoftver, finetime, corstime,
        jd, date, exptime, dpc_obs_id, start, instrmnt, roisizex,
        roisizey, roistrtx, roistrty, next, astrometry,
        targ_id, ra, dec, naxis1, naxis2, naxis3, naxis4, 
        badchecksum, baddatasum, filesize)
        VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)"""
        self.conn.commit()

        os.chmod(
            self.db_path,
            stat.S_IRUSR
            | stat.S_IWUSR  # owner: read/write
            | stat.S_IRGRP
            | stat.S_IWGRP,  # group: read/write
        )
        self.cur.execute(f"ATTACH DATABASE '{LEVEL0_DIR}/pointings.db' AS level0")
        self.cur.execute(f"ATTACH DATABASE '{LEVEL1_DIR}/level1.db' AS level1")
        self.cur.execute(f"ATTACH DATABASE '{LEVEL2_DIR}/level2.db' AS level2")

    def get_output_filename(self, filename_or_row):
        if isinstance(filename_or_row, tuple):
            fname = filename_or_row[0]
        elif isinstance(filename_or_row, str):
            filename = filename_or_row
            fname = filename.split("/")[-1] if "/" in filename else filename
        elif filename_or_row is None:
            return None
        self.cur.execute(
            f"SELECT targ_id, jd, ra, dec FROM level{self.level - 1}.pointings WHERE jd = (SELECT start FROM level{self.level - 1}.pointings WHERE filename=?)",
            (fname,),
        )
        row = self.cur.fetchone()
        if row is None:
            return None
        t = Time(row[1], format="jd").to_datetime()
        return f"{self.level_dir}/{t.year}/{t.month}/{t.day}/{get_dpc_hashkey(row[0], row[2], row[3])}/{Time(row[1], format='jd').strftime('%Y-%m-%d__%H-%M-%S')}_{row[0]}_v{__version__.replace('.', '-')}_l3.fits"

    def process(self, filename):
        logger.info(f"Processing {filename} to Level {self.level}.")
        logger.info("Ensuring file directory present.")
        path = self.get_output_filename(filename)
        os.makedirs("/".join(path.split("/")[:-1]), exist_ok=True)
        # Somehow write data to the level 3 file here...
        # if "VisSci" in filename:
        #     with VISDALevel0HDUList(filename) as hdulist:
        #         hdulist = getattr(hdulist, f"to_level{self.level}")()
        #         hdulist.writeto(path, overwrite=True, checksum=True)
        # if "InfImg" in filename:
        #     with NIRDALevel0HDUList(filename) as hdulist:
        #         hdulist = getattr(hdulist, f"to_level{self.level}")()
        #         hdulist.writeto(path, overwrite=True, checksum=True)
        if "VisImg" in filename:
            # VisImg are full frame images, they don't go into our level 3 products.
            logger.warning("Can not process VisImg to Level 3. Will skip this file.")
            return None
        logger.info(f"Wrote {filename.split('/')[-1]} to {path}")
        return self.get_entry(filename)



# flake8: noqa W291
"""Tools for keeping a database of pandora files"""

import os
import sqlite3
import stat
from concurrent.futures import ThreadPoolExecutor, as_completed
from copy import deepcopy

import numpy as np
from astropy.time import Time

from .. import CRSOFTVER, LEVEL0_DIR, LEVEL1_DIR, __version__, logger
from ..nirda import NIRDALevel0HDUList
from ..utils import get_dpc_hashkey
from ..visda import VISDAFFILevel0HDUList, VISDALevel0HDUList
from .mixins import DataBaseMixins


class Level1DataBase(DataBaseMixins):
    """Database for managing Level 1 files."""

    def __init__(self):
        self.level = 1
        self.level_dir = LEVEL1_DIR
        self.db_path = f"{self.level_dir}/level{self.level}.db"
        self.conn = sqlite3.connect(self.db_path, timeout=120)
        self.cur = self.conn.cursor()
        self.cur.execute(
            """
        CREATE TABLE IF NOT EXISTS pointings (
            filename TEXT PRIMARY KEY,
            lvlfilename TEXT,
            dir TEXT,
            crsoftver TEXT,
            pfsoftver TEXT,
            finetime INT,
            corstime INT,
            jd FLOAT,
            date STR,
            exptime FLOAT,
            dpc_obs_id INT,
            start FLOAT,
            instrmnt TEXT,
            roisizex INT,
            roisizey INT,
            roistrtx INT,
            roistrty INT,
            next INT,
            astrometry BOOL,
            targ_id STR,
            ra FLOAT,
            dec FLOAT,
            naxis1 INT,
            naxis2 INT,
            naxis3 INT,
            naxis4 INT,
            badchecksum INT,
            baddatasum INT,
            filesize FLOAT
        )
        """
        )
        self.update_str = """INSERT INTO pointings 
        (filename, lvlfilename, dir, crsoftver, pfsoftver, finetime, corstime,
        jd, date, exptime, dpc_obs_id, start, instrmnt, roisizex,
        roisizey, roistrtx, roistrty, next, astrometry,
        targ_id, ra, dec, naxis1, naxis2, naxis3, naxis4, 
        badchecksum, baddatasum, filesize)
        VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)"""
        self.conn.commit()

        os.chmod(
            self.db_path,
            stat.S_IRUSR
            | stat.S_IWUSR  # owner: read/write
            | stat.S_IRGRP
            | stat.S_IWGRP,  # group: read/write
        )
        # Add the level0 database in
        self.cur.execute(f"ATTACH DATABASE '{LEVEL0_DIR}/pointings.db' AS level0")

    def __repr__(self):
        return f"Pandora Level{self.level}DataBase"

    @property
    def n_files_to_process(self):
        self.cur.execute(
            f"""SELECT COUNT()
                FROM level{self.level - 1}.pointings AS src
                LEFT JOIN pointings AS dst
                        ON src.filename = dst.filename
                WHERE (src.crsoftver = ? AND src.badchecksum = 0 AND src.baddatasum = 0)
                    AND (dst.filename IS NULL OR dst.pfsoftver != ?);""",
            (CRSOFTVER, __version__),
        )
        nrows = self.cur.fetchone()[0]
        return nrows

    def get_files_to_process(self, crsoftver=CRSOFTVER, nchunks=None, chunk=None):
        # Compute limit/offset if chunking is requested
        if nchunks is not None and chunk is not None:
            # total files
            N = self.n_files_to_process
            chunk_size = int(np.ceil(N / nchunks))
            offset = chunk * chunk_size
            limit = chunk_size
        else:
            limit = None
            offset = None

        base_sql = f"""
            SELECT src.dir, src.filename
            FROM level{self.level - 1}.pointings AS src
            LEFT JOIN pointings AS dst
                 ON src.filename = dst.filename
            WHERE (src.crsoftver = ? AND src.badchecksum = 0 AND src.baddatasum = 0)
              AND (dst.filename IS NULL OR dst.pfsoftver != ?)
        """

        params = [crsoftver, __version__]

        # Add LIMIT/OFFSET only if chunking
        if limit is not None:
            base_sql += " LIMIT ? OFFSET ?"
            params.extend([limit, offset])

        self.cur.execute(base_sql, tuple(params))
        rows = self.cur.fetchall()

        if rows:
            return [f"{d}/{f}" for d, f in rows]
        else:
            return []

    # def _is_bad_value(
    #     self, fname, expected_entry, entry, name, level, remove_bad_entries=True
    # ):
    #     if expected_entry != entry:
    #         logger.warning(
    #             f"File {fname} exists in Level {level} database with `{name}` set to `{entry}` requested value was `{expected_entry}`."
    #         )
    #         if remove_bad_entries:
    #             if level == 0:
    #                 self.cur.execute(
    #                     "DELETE FROM level0.pointings WHERE filename=?", (fname,)
    #                 )
    #             if level == self.level:
    #                 self.cur.execute("DELETE FROM pointings WHERE filename=?", (fname,))
    #             else:
    #                 raise ValueError(f"Can not parse usage of level `{level}`")
    #             self.conn.commit()
    #             logger.warning(f"File {fname} removed from Level {level} database.")
    #         return True
    #     return False

    # def _is_bad_row(
    #     self,
    #     filename,
    #     crsoftver=CRSOFTVER,
    #     pfsoftver=__version__,
    #     badchecksum=0,
    #     baddatasum=0,
    #     level=0,
    #     remove_bad_entries=False,
    # ):
    #     fname = filename.split("/")[-1] if "/" in filename else filename
    #     self.cur.execute(
    #         f"SELECT crsoftver, pfsoftver, badchecksum, baddatasum FROM {'level0.' if level == 0 else ''}pointings WHERE filename=?",
    #         (fname,),
    #     )
    #     row = self.cur.fetchone()

    #     if row is None:
    #         return False
    #     (
    #         database_crsoftver,
    #         database_pfsoftver,
    #         database_badchecksum,
    #         database_baddatasum,
    #     ) = row
    #     is_bad_row = False
    #     # Log any discrepancies
    #     if crsoftver is not None:
    #         is_bad_row |= self._is_bad_value(
    #             fname=fname,
    #             expected_entry=crsoftver,
    #             entry=database_crsoftver,
    #             name="CRSOFTV",
    #             level=0,
    #             remove_bad_entries=remove_bad_entries,
    #         )
    #     if pfsoftver is not None:
    #         if database_pfsoftver is not None:
    #             is_bad_row |= self._is_bad_value(
    #                 fname=fname,
    #                 expected_entry=pfsoftver,
    #                 entry=database_pfsoftver,
    #                 name="PFSOFTV",
    #                 level=0,
    #                 remove_bad_entries=remove_bad_entries,
    #             )
    #     if badchecksum is not None:
    #         is_bad_row |= self._is_bad_value(
    #             fname=fname,
    #             expected_entry=badchecksum,
    #             entry=database_badchecksum,
    #             name="badchecksum",
    #             level=0,
    #             remove_bad_entries=remove_bad_entries,
    #         )
    #     if baddatasum is not None:
    #         is_bad_row |= self._is_bad_value(
    #             fname=fname,
    #             expected_entry=baddatasum,
    #             entry=database_baddatasum,
    #             name="baddatasum",
    #             level=0,
    #             remove_bad_entries=remove_bad_entries,
    #         )
    #     return is_bad_row

    def check_filename_in_database(self, filename, level=0):
        fname = filename.split("/")[-1] if "/" in filename else filename
        self.cur.execute(
            f"SELECT pfsoftver FROM {f'level{self.level - 1}.' if level == 0 else ''}pointings WHERE filename=?",
            (fname,),
        )
        return self.cur.fetchone() is not None

    # def files_to_process(self, crsoftver=CRSOFTVER, nchunks=1, chunknumber=1):
    #     cur = self.conn.cursor()
    #     cur.execute("SELECT filename, dir FROM level0.pointings")

    #     row = cur.fetchone()
    #     while row is not None:
    #         if (not self.check_filename_in_database(row[0], level=self.level)) & (
    #             not self._is_bad_row(row[0], crsoftver=crsoftver, level=0)
    #         ):
    #             yield (f"{row[1]}/{row[0]}")
    #         row = cur.fetchone()

    def get_entry(self, filename):
        fname = filename.split("/")[-1] if "/" in filename else filename
        self.cur.execute(
            "SELECT * FROM level0.pointings WHERE filename=?",
            (fname,),
        )
        row = self.cur.fetchone()
        return (row[0], self.get_output_filename(row), *row[1:])

    def get_output_filename(self, filename_or_row):
        if isinstance(filename_or_row, tuple):
            row = filename_or_row
            t = Time(row[10], format="jd").to_datetime()
            targ_id, ra, dec = row[18], row[19], row[20]
            fname = row[0]
        elif isinstance(filename_or_row, str):
            filename = filename_or_row
            fname = filename.split("/")[-1] if "/" in filename else filename
            self.cur.execute(
                f"SELECT start, targ_id, ra, dec FROM level{self.level - 1}.pointings WHERE filename=?",
                (fname,),
            )
            row = self.cur.fetchone()
            t = Time(row[0], format="jd").to_datetime()
            targ_id, ra, dec = row[1:]
            if row is None:
                return None
        elif filename_or_row is None:
            return None
        fname_no_suffix = ".".join(fname.split(".")[:-1])
        suffix = fname.split(".")[-1]
        return f"{self.level_dir}/{t.year}/{t.month}/{t.day}/{get_dpc_hashkey(targ_id, ra, dec)}/{fname_no_suffix}_v{__version__.replace('.', '-')}_l{self.level}.{suffix}"

    def process(self, filename):
        logger.info(f"Processing {filename} to Level {self.level}.")
        logger.info("Ensuring file directory present.")
        path = self.get_output_filename(filename)
        os.makedirs("/".join(path.split("/")[:-1]), exist_ok=True)
        if "VisSci" in filename:
            with VISDALevel0HDUList(filename) as hdulist:
                hdulist = getattr(hdulist, f"to_level{self.level}")()
                hdulist.writeto(path, overwrite=True, checksum=True)
        if "VisImg" in filename:
            with VISDAFFILevel0HDUList(filename) as hdulist:
                hdulist = getattr(hdulist, f"to_level{self.level}")()
                hdulist.writeto(path, overwrite=True, checksum=True)
        if "InfImg" in filename:
            with NIRDALevel0HDUList(filename) as hdulist:
                hdulist = getattr(hdulist, f"to_level{self.level}")()
                hdulist.writeto(path, overwrite=True, checksum=True)
        logger.info(f"Wrote {filename.split('/')[-1]} to {path}")
        return self.get_entry(filename)

    # def crawl_and_process_parallel(self, max_workers=16, crsoftver=CRSOFTVER):
    #     paths = list(self.files_to_process(crsoftver=crsoftver))
    #     for path in paths:
    #         out_path = self.get_output_filename(path)
    #         os.makedirs("/".join(out_path.split("/")[:-1]), exist_ok=True)
    #     rows = []
    #     with ThreadPoolExecutor(max_workers=max_workers) as ex:
    #         futures = [ex.submit(self.process, p) for p in paths]
    #         for fut in as_completed(futures):
    #             row = fut.result()
    #             if row is not None:
    #                 rows.append(row)
    #     self.add_entries(rows)

    def crawl_and_process(self, crsoftver=CRSOFTVER, nchunks=None, chunk=None):
        for path in self.get_files_to_process(
            crsoftver=crsoftver, nchunks=nchunks, chunk=chunk
        ):
            self.add_entry(self.process(path))

    # def crawl_and_process_parallel(self, max_workers=16, crsoftver=CRSOFTVER):
    #     paths = list(self.files_to_process(crsoftver=crsoftver))
    #     for path in paths:
    #         out_path = self.get_output_filename(path)
    #         os.makedirs("/".join(out_path.split("/")[:-1]), exist_ok=True)
    #     rows = []
    #     with ThreadPoolExecutor(max_workers=max_workers) as ex:
    #         futures = [ex.submit(self.process, p) for p in paths]
    #         for fut in as_completed(futures):
    #             row = fut.result()
    #             if row is not None:
    #                 rows.append(row)
    #     self.add_entries(rows)



"""Functions for processing data"""

from datetime import timedelta

import astropy.units as u
import numpy as np
from astropy.io import fits
from astropy.time import Time


class ProcessingMixins:
    @property
    def start_time(self):
        """Given Pandora HDUList obtains the detector time in TAI."""
        time = (
            Time("2000-01-01 12:00:00", scale="tai")
            + timedelta(
                seconds=self[0].header["CORSTIME"],
                milliseconds=self[0].header["FINETIME"] / 1e6,
            )
        ).utc
        return time

    @property
    def pixel_coordinates(self):
        hdr = self[0].header
        R, C = np.mgrid[
            hdr["ROISTRTX"] : hdr["ROISTRTX"] + hdr["ROISIZEX"],
            hdr["ROISTRTY"] : hdr["ROISTRTY"] + hdr["ROISIZEY"],
        ]
        return R, C

    @property
    def row(self):
        hdr = self[0].header
        return np.arange(hdr["ROISTRTX"], hdr["ROISTRTX"] + hdr["ROISIZEX"])

    @property
    def column(self):
        hdr = self[0].header
        return np.arange(hdr["ROISTRTY"], hdr["ROISTRTY"] + hdr["ROISIZEY"])

    @property
    def frame_time(self):
        if "FRMTIME" in self[0].header:
            return (u.millisecond * self[0].header["FRMTIME"]).to(u.second)
        elif "EXPTIMEU" in self[0].header:
            return (
                u.microsecond
                * self[0].header["EXPTIMEU"]
                * self[0].header["FRMSCLCT"]
                / self[1].header["NAXIS3"]
            ).to(u.second)
        elif "EXPTIME" in self[0].header:
            return (
                u.microsecond
                * self[0].header["EXPTIME"]
                * self[0].header["FRMSCLCT"]
                / self[1].header["NAXIS3"]
            ).to(u.second)

    @property
    def nframes(self):
        return self[1].header[f"NAXIS{self[1].header['NAXIS']}"]

    @property
    def end_time(self):
        return self.start_time + self.nframes * self.frame_time

    def apply_quality(self):
        quality = self.reference.get_bad_pixel()[self.row][:, self.column]
        hdu = fits.CompImageHDU(quality, name="QUALITY")
        self.append(hdu)
        self[0].header["COMMENT"] = "Data quality extension created."



"""Mixins for databases"""

import os
from datetime import datetime, timezone

import numpy as np
import pandas as pd
from astropy.time import Time

from .. import LEVEL1_DIR, LEVEL2_DIR, LEVEL3_DIR, logger  # noqa


def _process_time(time):
    if isinstance(time, Time):
        time = time.jd
    elif not isinstance(time, float):
        try:
            time = Time(time).jd
        except ValueError:
            raise ValueError("`time_range` must be in JD.")
    return time


class DataBaseMixins:
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        """Closes the database connection."""
        if self.conn:
            self.conn.close()

    def to_pandas(self, time_range=None, targ_id=None, dpc_obs_id=None):
        sql = "SELECT * FROM pointings"
        params = []

        where_clauses = []

        if time_range is not None:
            start, end = _process_time(time_range[0]), _process_time(time_range[1])
            start, end = np.sort([start, end])
            where_clauses.append("jd BETWEEN ? AND ?")
            params.extend([start, end])

        if targ_id is not None:
            where_clauses.append("targ_id = ?")
            params.append(targ_id)

        if dpc_obs_id is not None:
            where_clauses.append("dpc_obs_id = ?")
            params.append(dpc_obs_id)

        if where_clauses:
            sql += " WHERE " + " AND ".join(where_clauses)

        return pd.read_sql_query(sql, self.conn, params=params)

    def to_archive_manifest(self):
        logger.info(f"Creating Level {self.level} archive manifest.")
        manifest_path = (
            globals()[f"LEVEL{self.level}_DIR"]
            + "/"
            + f"level{self.level}_manifest.csv"
        )
        columns = [
            "Target Name",
            "RA",
            "Dec",
            "Obs. Date Start UT",
            "Obs. Date Start",
            "Obs. Date End UT",
            "Obs. Date End",
            "Level of file",
            "Detector",
            "Processing version",
            "Full file path",
            "Filename",
            "Processing Date",
            "Delivery Date",
        ]
        df = self.to_pandas()
        k = np.asarray([os.path.isfile(path) for path in df.lvlfilename.values])
        if not k.any():
            logger.info(
                f"No files found for level {self.level} archive manifest. Storing at {manifest_path}"
            )
            return pd.DataFrame(columns=columns)
        adf = (
            df[k][
                [
                    "targ_id",
                    "ra",
                    "dec",
                    "jd",
                    "instrmnt",
                    "pfsoftver",
                    "lvlfilename",
                    "exptime",
                ]
            ]
            .copy()
            .reset_index(drop=True)
        )
        adf.loc[:, "Obs. Date Start UT"] = Time(adf.jd.values, format="jd").isot
        adf.loc[:, "Obs. Date End"] = (
            adf["jd"].values.copy() + adf.exptime.values.copy() / 86400.0
        )
        adf.loc[:, "Obs. Date End UT"] = Time(
            adf["Obs. Date End"].values.copy(), format="jd"
        ).isot
        adf["Level of file"] = self.level
        adf["Processing Date"] = [
            datetime.fromtimestamp(os.path.getmtime(path), tz=timezone.utc)
            .replace(tzinfo=None)
            .isoformat(timespec="milliseconds")
            for path in adf["lvlfilename"]
        ]
        adf["Delivery Date"] = Time.now().isot
        adf["Filename"] = [path.split("/")[-1] for path in adf.lvlfilename.values]
        adf = adf.rename(
            {
                "targ_id": "Target Name",
                "ra": "RA",
                "dec": "Dec",
                "jd": "Obs. Date Start",
                "instrmnt": "Detector",
                "pfsoftver": "Processing version",
                "lvlfilename": "Full file path",
            },
            axis="columns",
        )
        adf = adf[columns]
        logger.info(f"Archive manifest stored at {manifest_path}")

        return adf

    def add_entry(self, values):
        if values is not None:
            self.cur.execute(
                self.update_str,
                values,
            )
            self.conn.commit()

    def add_entries(self, values):
        self.cur.executemany(
            self.update_str,
            [v for v in values if v is not None],
        )
        self.conn.commit()



import os
import shutil

import numpy as np
import pandas as pd

from .. import (
    CRSOFTVER,
    DATA_DIR,
    LEVEL0_DIR,
    LEVEL1_DIR,
    LEVEL2_DIR,
    LEVEL3_DIR,
    __version__,
    logger,
)
from . import Level0DataBase, Level1DataBase, Level2DataBase, Level3DataBase  # noqa

__all__ = [
    "update_filedatabase",
    "delete_filedatabase",
    "delete_level1database",
    "delete_level2database",
    "delete_level1filestorage",
    "delete_level2filestorage",
    "delete_all",
    "get_status",
]


def update_filedatabase() -> None:
    """
    Creates and updates to the SQLite database file.
    """
    with Level0DataBase() as db:
        db.crawl_and_add_parallel(DATA_DIR)
        db.update_pointings()


def delete_filedatabase() -> None:
    """
    Deletes the SQLite database file.

    Raises
    ------
    FileNotFoundError
        If the database file does not exist.
    """
    db_path = f"{LEVEL0_DIR}/pointings.db"
    if os.path.exists(db_path):
        os.remove(db_path)
        logger.warning(f"Database at {db_path} has been deleted.")
    else:
        logger.warning(
            f"Tried to delete Level0DataBase. No database found at {db_path}."
        )


def delete_level1database() -> None:
    """
    Deletes the SQLite database file.

    Raises
    ------
    FileNotFoundError
        If the database file does not exist.
    """
    db_path = f"{LEVEL1_DIR}/level1.db"
    if os.path.exists(db_path):
        os.remove(db_path)
        logger.warning(f"Database at {db_path} has been deleted.")
    else:
        logger.warning(
            f"Tried to delete Level1DataBase. No database found at {db_path}."
        )


def delete_level1filestorage():
    logger.info("Running `delete_level1filestorage`")
    if os.path.exists(LEVEL1_DIR):
        shutil.rmtree(LEVEL1_DIR)
    if not os.path.exists(LEVEL1_DIR):
        os.makedirs(LEVEL1_DIR, exist_ok=True)
        os.chmod(LEVEL1_DIR, 0o750)
    logger.info("Finished `delete_level1filestorage`")


def delete_level2database() -> None:
    """
    Deletes the SQLite database file.

    Raises
    ------
    FileNotFoundError
        If the database file does not exist.
    """
    db_path = f"{LEVEL2_DIR}/level2.db"
    if os.path.exists(db_path):
        os.remove(db_path)
        logger.warning(f"Database at {db_path} has been deleted.")
    else:
        logger.warning(
            f"Tried to delete Level2DataBase. No database found at {db_path}."
        )


def delete_level2filestorage():
    logger.info("Running `delete_level2filestorage`")
    if os.path.exists(LEVEL2_DIR):
        shutil.rmtree(LEVEL2_DIR)
    if not os.path.exists(LEVEL2_DIR):
        os.makedirs(LEVEL2_DIR, exist_ok=True)
        os.chmod(LEVEL2_DIR, 0o750)
    logger.info("Finished `delete_level2filestorage`")


def delete_level3database() -> None:
    """
    Deletes the SQLite database file.

    Raises
    ------
    FileNotFoundError
        If the database file does not exist.
    """
    db_path = f"{LEVEL3_DIR}/level3.db"
    if os.path.exists(db_path):
        os.remove(db_path)
        logger.warning(f"Database at {db_path} has been deleted.")
    else:
        logger.warning(
            f"Tried to delete Level3DataBase. No database found at {db_path}."
        )


def delete_level3filestorage():
    logger.info("Running `delete_level3filestorage`")
    if os.path.exists(LEVEL3_DIR):
        shutil.rmtree(LEVEL3_DIR)
    if not os.path.exists(LEVEL3_DIR):
        os.makedirs(LEVEL3_DIR, exist_ok=True)
        os.chmod(LEVEL3_DIR, 0o750)
    logger.info("Finished `delete_level3filestorage`")


def delete_all():
    """Clear all processing data"""
    delete_filedatabase()
    delete_level1database()
    delete_level1filestorage()
    delete_level2filestorage()
    delete_level2filestorage()
    delete_level3filestorage()
    delete_level3filestorage()


def get_status():
    def get_nrows(level):
        with globals()[f"Level{level}DataBase"]() as self:
            self.cur.execute("""SELECT COUNT() FROM pointings""")
            nrows = self.cur.fetchone()[0]
        return nrows

    def get_ntargets(level):
        with globals()[f"Level{level}DataBase"]() as self:
            self.cur.execute("""SELECT COUNT(DISTINCT targ_id) FROM pointings""")
            nrows = self.cur.fetchone()[0]
        return nrows

    def get_npointings(level):
        with globals()[f"Level{level}DataBase"]() as self:
            self.cur.execute("""SELECT COUNT(DISTINCT start) FROM pointings""")
            nrows = self.cur.fetchone()[0]
        return nrows

    def get_nbadcrsoftver(level):
        with globals()[f"Level{level}DataBase"]() as self:
            self.cur.execute(
                """SELECT COUNT() FROM pointings WHERE crsoftver != ?""",
                (CRSOFTVER,),
            )
            nrows = self.cur.fetchone()[0]
        return nrows

    def get_nbadpfsoftver(level):
        with globals()[f"Level{level}DataBase"]() as self:
            self.cur.execute(
                """SELECT COUNT() FROM pointings WHERE pfsoftver != ?""",
                (__version__,),
            )
            nrows = self.cur.fetchone()[0]
        return nrows

    def get_nbadchecksum(level):
        with globals()[f"Level{level}DataBase"]() as self:
            self.cur.execute(
                """SELECT COUNT() FROM pointings WHERE badchecksum != ?""", (0,)
            )
            nrows = self.cur.fetchone()[0]
        return nrows

    def get_nbaddatasum(level):
        with globals()[f"Level{level}DataBase"]() as self:
            self.cur.execute(
                """SELECT COUNT() FROM pointings WHERE baddatasum != ?""", (0,)
            )
            nrows = self.cur.fetchone()[0]
        return nrows

    r = {
        "n_row_s": [get_nrows(level) for level in np.arange(4)],
        "n_bad_crsoftver": [get_nbadcrsoftver(level) for level in np.arange(4)],
        "n_bad_pfsoftver": [get_nbadpfsoftver(level) for level in np.arange(4)],
        "n_bad_checksum": [get_nbadchecksum(level) for level in np.arange(4)],
        "n_bad_datasum": [get_nbaddatasum(level) for level in np.arange(4)],
        "n_unique_targets": [get_ntargets(level) for level in np.arange(4)],
        "n_unique_pointings": [get_npointings(level) for level in np.arange(4)],
    }
    return pd.DataFrame.from_dict(r).T



import numpy as np
from astropy.coordinates import SkyCoord
from astropy.table import Table

from . import FORMATSDIR, VISDAReference
from .fits import PandoraHDUList

__all__ = [
    "VISDAFFILevel0HDUList",
    "VISDAFFILevel1HDUList",
    "VISDALevel0HDUList",
    "VISDALevel1HDUList",
    "VISDALevel2HDUList",
]


class VISDAFFILevel0HDUList(PandoraHDUList):
    filename = FORMATSDIR + "visda/level0-ffi_visda.xlsx"
    reference = VISDAReference

    def to_level1(self):
        return VISDAFFILevel1HDUList(self)


class VISDAFFILevel1HDUList(VISDAFFILevel0HDUList):
    filename = FORMATSDIR + "visda/level1-ffi_visda.xlsx"
    reference = VISDAReference

    def to_level1(self):
        raise ValueError("This is a level 1 product.")


class VISDALevel0HDUList(PandoraHDUList):
    filename = FORMATSDIR + "visda/level0_visda.xlsx"
    reference = VISDAReference

    @property
    def coord(self):
        return SkyCoord(
            self[0].header["TARG_RA"], self[0].header["TARG_DEC"], unit="deg"
        )

    @property
    def ROI_corners(self):
        roistrtx = self[0].header["ROISTRTX"]
        roistrty = self[0].header["ROISTRTY"]
        return [
            (y + roistrty, x + roistrtx)
            for x, y in Table(self[2].data).to_pandas().values
        ]

    @property
    def ROI_size(self):
        return (self[0].header["STARDIMS"], self[0].header["STARDIMS"])

    @property
    def nROI(self):
        return self[0].header["NUMSTARS"]

    @property
    def data_cube(self):
        numSubFrms = int(np.ceil(np.sqrt(self.nROI)))
        dims = (numSubFrms * self.ROI_size[0], numSubFrms * self.ROI_size[1])
        if self[1].header["NAXIS1"] == dims[0]:
            ex = 0
        elif self[1].header["NAXIS1"] == (dims[0] + numSubFrms + 1):
            ex = 1
        else:
            raise ValueError("Can not parse the padding dimensions.")
        data = self[1].data[:, ex:, ex:]
        shape = data.shape[1:]
        width = self.ROI_size[0]
        nims = int(shape[1] / (width + ex))
        d1 = np.asarray(np.array_split(data, nims, axis=2))
        nims = int(shape[0] / (width + ex))
        d2 = np.asarray(np.array_split(d1, nims, axis=2))[
            :, :, :, :-ex, :-ex
        ].transpose([2, 0, 1, 3, 4])
        return d2

    @property
    def star_list(self):
        d = self.data_cube[:, ::-1]
        starlist = []
        numSubFrms = int(np.ceil(np.sqrt(self.nROI)))
        for idx in range(numSubFrms):
            for jdx in range(numSubFrms):
                if (idx * numSubFrms + jdx) == self.nROI:
                    return np.asarray(starlist).transpose([1, 0, 2, 3])
                starlist.append(d[:, idx, jdx])

    @property
    def star_row(self):
        return np.asarray(
            [r + np.arange(self.ROI_size[0]) for r, _ in self.ROI_corners]
        )

    @property
    def star_column(self):
        return np.asarray(
            [c + np.arange(self.ROI_size[1]) for _, c in self.ROI_corners]
        )

    def to_level1(self):
        return VISDALevel1HDUList(self)


class VISDALevel1HDUList(VISDALevel0HDUList):
    filename = FORMATSDIR + "visda/level1_visda.xlsx"
    reference = VISDAReference

    def to_level1(self):
        raise ValueError("This is a level 1 product.")


class VISDALevel2HDUList(PandoraHDUList):
    filename = FORMATSDIR + "visda/level2_visda.xlsx"
    reference = VISDAReference
